/* 
  err_mac - convenient macros for error-checking
 
  Copyright (C) 2007-2009 William Trevor King
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of the
  License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
 
  The author may be contacted at <wking@drexel.edu> on the Internet, or
  write to Trevor King, Drexel University, Physics Dept., 3141 Chestnut St.,
  Philadelphia PA 19104, USA.
*/

#ifndef _ERR_MAC_H
#define _ERR_MAC_H

#include <stdio.h> /* for fprintf(), stderr */

#define SUCCESS 0
#define FAILURE 1

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
static int err_mac_ans;

/*
 * For functions returning int, when you want an immediate return.
 */

/* CHK() fails if (a != SUCCESS) */
#define CHK(a)                                                        \
  do {                                                                \
    if( (err_mac_ans = (a)) != SUCCESS ) {                            \
      fprintf(stderr,"FAILURE: %s.%d. (%s) returned %d\n",__FILE__,__LINE__, #a, err_mac_ans); \
      fflush(stderr);                                                 \
      return err_mac_ans;                                             \
    }                                                                 \
  } while (0)

/* AST() fails if (a != 0) */
#define AST(a)                                                        \
  do {                                                                \
    if( (err_mac_ans = (a)) == FALSE ) {                              \
      fprintf(stderr, "ASSERT FAILED: %s.%d (%s) returned %d\n",__FILE__,__LINE__, #a, err_mac_ans); \
      fflush(stderr);                                                 \
      return FAILURE;                                                 \
    }                                                                 \
  } while(0)

/* like AST(), but return E_CODE instead of the default FAILURE */
#define E_AST(a,e_code)                                               \
  do {                                                                \
    if((a) == FALSE) {                                                \
      fprintf(stderr,"ASSERT FAILED: %s.%d (%s)\n",__FILE__,__LINE__, #a); \
      fflush(stderr);                                                 \
      return (e_code);                                                \
    }                                                                 \
  } while(0)

/* like E_AST(), but also fprintfs __VA_ARGS__ */
#define M_AST(a,e_code,...)                                           \
  do {                                                                \
    if((a) == FALSE) {                                                \
      fprintf(stderr,"ASSERT FAILED: %s.%d (%s) ",__FILE__,__LINE__, #a); \
      fprintf(stderr, __VA_ARGS__);                                   \
      fflush(stderr);                                                 \
      return (e_code);                                                \
    }                                                                 \
  } while(0)

/* print a message and return E_CODE */
#define M_EXIT(e_code, ... )                               \
  do {                                                     \
    fprintf(stderr,"EXIT: %s.%d ",__FILE__,__LINE__);      \
    fprintf(stderr, __VA_ARGS__ );                         \
    fflush(stderr);                                        \
    return (e_code);                                       \
  } while(0)

/*
 * When you don't want an immedate return, but would rather 'goto Error;'.
 * These set the local variable RC with the return value before jumping.
 */

#define G_CHK(a)                                                      \
  do {                                                                \
    if( (rc = (a)) != SUCCESS ) {                                     \
      fprintf(stderr,"FAILURE: %s.%d. (%s)\n",__FILE__,__LINE__, #a); \
      fflush(stderr);                                                 \
      goto Error;                                                     \
    }                                                                 \
  } while (0)

#define G_AST(a)                                                      \
  do {                                                                \
    if( (rc = (a)) == FALSE ) {                                       \
      fprintf(stderr, "ASSERT FAILED: %s.%d (%s)\n",__FILE__,__LINE__, #a); \
      fflush(stderr);                                                 \
      goto Error;                                                     \
    }                                                                 \
  } while(0)

#define G_E_AST(a,e_code)                                             \
  do {                                                                \
    if((a) == FALSE) {                                                \
      fprintf(stderr,"ASSERT FAILED: %s.%d (%s)\n",__FILE__,__LINE__, #a); \
      fflush(stderr);                                                 \
      rc = e_code;                                                    \
      goto Error;                                                     \
    }                                                                 \
  } while(0)

#define G_M_AST(a,e_code,...)                                         \
  do {                                                                \
    if((a) == FALSE) {                                                \
      fprintf(stderr,"ASSERT FAILED: %s.%d (%s) ",__FILE__,__LINE__, #a); \
      fprintf(stderr, __VA_ARGS__);                                   \
      fflush(stderr);                                                 \
      rc = e_code;                                                    \
      goto Error;                                                     \
    }                                                                 \
  } while(0)

/* print a message and return E_CODE */
#define G_M_EXIT(e_code, ... )                             \
  do {                                                     \
    fprintf(stderr,"EXIT: %s.%d ",__FILE__,__LINE__);      \
    fprintf(stderr, __VA_ARGS__ );                         \
    fflush(stderr);                                        \
    rc = e_code;                                           \
    goto Error;                                            \
  } while(0)

/*
 * And when you only want to print a message
 */
#define P_CHK(a)                                                      \
  do {                                                                \
    if( (a) != SUCCESS ) {                                            \
      fprintf(stderr,"FAILURE: %s.%d. (%s)\n",__FILE__,__LINE__, #a); \
      fflush(stderr);                                                 \
    }                                                                 \
  } while (0)

/* AST() fails if (a != 0) */
#define P_AST(a)                                                      \
  do {                                                                \
    if( (a) == FALSE ) {                                              \
      fprintf(stderr, "ASSERT FAILED: %s.%d (%s)\n",__FILE__,__LINE__, #a); \
      fflush(stderr);                                                 \
    }                                                                 \
  } while(0)

#endif /* _ERR_MAC_H */
