/*
  strip - program for stripcharting ASCII data read from a named pipe.
  Copyright (C) 2008, William Trevor King

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  The author may be contacted at <wking@drexel.edu> on the Internet, or
  write to Trevor King, Drexel University, Physics Dept., 3141 Chestnut St.,
  Philadelphia PA 19104, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/stat.h> /* mkfifo(), S_IRUSR, S_IWUSR */
#include "err_mac.h"
#include "stripchart.h"

#define BUFSZ 81

//#define AUTOGEN /* for testing */


static int string2double(char *string, double *pDouble);
static int getargs(int argc, char **argv, unsigned int *pFlags,
		   char **pTitle, char **pGeometry, char **pPipe_name,
		   double *pY_max, double *pY_min, int *pPoints_shown_at_once);
static int open_pipe(char *name, unsigned int flags, FILE **pInStream);
static int close_pipe(char *name, FILE **pInStream);
static int get_num(FILE *stream, double *px);

static const char options[] = "t:g:p:cuM:m:n:avh";
static const char *func_name = NULL;
static const char *description = 
 "Pops up an X window strip chart, and a FIFO to dump data into";

static int verbose = 0;

#define FLAG_CREATE_FIFO 0x1
#define FLAG_AXES_ON 0x2

static void help(char *title, char *geometry, char *pipe_name,
		 double y_max, double y_min, int num_points)
{
  printf("usage: %s [-%s]\n", func_name, options);
  puts(description);
  printf("Version: %s\n\n", VERSION);

  puts("Options:");

  printf("-t <title>\tSet the window title (default: %s)\n", title);
  printf("-g <geometry>\tSet the initial window geometry (default: %s)\n", geometry);
  printf("-p <file>\tSet FIFO name (default: %s)\n", pipe_name);

  printf(" %s can either create its own FIFO or use an already existing one.\n",
	 func_name);
  puts(" Using an existing FIFO avoids the following race condition:");
  puts("\tFork");
  printf("\tProcess 1 starts %s\n", func_name);
  puts("\tProcess 2 sends data to FIFO, but it doesn't exist,");
  puts("\t\tso file system creates a new file");
  puts("\tProcess 1 gets around to trying to create the FIFO,");
  puts("\t\tbut fails because the file already exists");
  printf("-u\t%s uses an already created FIFO\n", func_name);
  printf("-c\t%s create the input FIFO (default)\n", func_name);

  printf("-M\tSet Maximum y value (default: %g)\n", y_max);
  printf("-m\tSet minimum y value (default: %g)\n", y_min);
  printf("-n\tSet number of points shown at once (default: %d)\n", num_points);
  puts("-a\tToggle axes (default: on)");

  puts("-v\tIncrement verbosity");
  puts("-h\tDisplay this help and exit");
  exit(1);
}

int main(int argc, char **argv)
{
  stripchart_t *chart = NULL;
  FILE *input;
  double x=0, y_max, y_min;
  unsigned int flags;
  int points_shown_at_once;
  char *title, *geometry, *pipe_name;

  CHK( getargs(argc, argv, &flags,
	       &title, &geometry, &pipe_name,
	       &y_max, &y_min, &points_shown_at_once) );
#ifndef AUTOGEN
  CHK( open_pipe(pipe_name, flags, &input) );
  if (verbose)
    fprintf(stderr, "creating stripchart on pipe %s with geom %s\n",
	    pipe_name, geometry);
#endif
  CHK( stripchart_create(&chart, title, geometry,
			 y_min, y_max, 1.0, 
			 points_shown_at_once,
			 flags & FLAG_AXES_ON) );
  do {
#ifndef AUTOGEN
    if (get_num(input, &x) != 0)
      break;
#else 
    x += 0.5;
    if (x == 9) break;
    usleep(100000);
#endif
    CHK( stripchart_point(chart, x) );
  } while (TRUE);
  CHK( stripchart_destroy(&chart) );
#ifndef AUTOGEN
  CHK( close_pipe(pipe_name, &input) );
#endif
  return 0;
}

static int string2double(char *string, double *pDouble)
{
  char *endp;
  if (errno != 0) {
    //perror("sloppy error checking");
    //fprintf(stderr, "errno = %d\n", errno);
    //return 1; /* unrecognized input */
    /* HACK!!!  Some Xlib function calls (XOpenDisplay, and others) seem to set
     * errno to 29, or 11, but still return successfully.  I have no idea why,
     * since I can't find any mention of errno in the Xlib documents, but we'll
     * just ignore any previous errnos and hope everything works out...
     */
    errno = 0;
  }
  *pDouble = strtod(string, &endp);
  if (errno != 0 || endp[0] != '\0') {
    if (errno != 0) {
      perror("strtod");
      fprintf(stderr, "could not parse '%s'\n", string);
      fprintf(stderr, "endp = \"%s\"\n", endp);
      fprintf(stderr, "errno = %d\n", errno);
    }
    return 1; /* unrecognized input */
  }
  return 0;
}

static int getargs(int argc, char **argv, unsigned int *pFlags,
		   char **pTitle, char **pGeometry, char **pPipe_name,
		   double *pY_max, double *pY_min, int *pPoints_shown_at_once)
{
  int c;
  double temp;

  func_name = argv[0];
  *pFlags = FLAG_CREATE_FIFO | FLAG_AXES_ON;
  *pTitle = "Stripchart";
  *pGeometry = "500x200+0+0";
  *pPipe_name = "strip_pipe";
  *pY_max = 10.0;
  *pY_min = -10.0;
  *pPoints_shown_at_once = 100;

  while ((c=getopt(argc, argv, options)) != -1) {
    switch (c) {
    case 't':
      *pTitle = optarg;
      break;
    case 'g':
      *pGeometry = optarg;
      break;
    case 'p':
      *pPipe_name = optarg;
      break;
    case 'u':
      *pFlags &= ~FLAG_CREATE_FIFO;
      break;
    case 'c':
      *pFlags |= FLAG_CREATE_FIFO;
      break;
    case 'M':
      if (string2double(optarg, pY_max) != 0)
	goto error;
      break;
    case 'm':
      if (string2double(optarg, pY_min) != 0)
	goto error;
      break;
    case 'n':
      if (string2double(optarg, &temp) != 0)
	goto error;
      *pPoints_shown_at_once = (int)temp;
      break;
    case 'a':
      *pFlags ^= FLAG_AXES_ON;
      break;
    case 'v':
      verbose++;
      break;
    default:
      printf("Unrecognized option '%c'\n", c);
    case 'h': /* default falls through */
    error:
      printf("Problem with option '%c' (%s)\n", c, optarg);
      help(*pTitle, *pGeometry, *pPipe_name, *pY_max, *pY_min, *pPoints_shown_at_once);
      break;
    }
  }
  return 0;
}

static int open_pipe(char *name, unsigned int flags, FILE **pStream)
{
  if (flags & FLAG_CREATE_FIFO) {
    /* make a fifo with read/write permision for the user */
    if (mkfifo(name, S_IRUSR | S_IWUSR) != 0) {
      perror("mkfifo");
      fprintf(stderr, "Error creating FIFO '%s', aborting\n", name);
      AST(1==0);
    }
  }
  AST( (*pStream = fopen(name, "a+")) != NULL );
  return 0;
}

static int close_pipe(char *name, FILE **pInStream)
{
  CHK( fclose(*pInStream) );
  CHK( remove( name ) );
  return 0;
}

static int get_num(FILE *stream, double *px)
{
  char buf[BUFSZ];
  
  if (fgets(buf, BUFSZ, stream) == NULL)
    return 1; /* end of file or error */
  if (buf[strlen(buf)-1] == '\n')
    buf[strlen(buf)-1] = '\0';
  if ( string2double(buf, px) != 0)
    return 1;
  return 0;
}
