/*
  stripchart - tools for quick & dirty data plotting.
  Copyright (C) 2008, William Trevor King

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  The author may be contacted at <wking@drexel.edu> on the Internet, or
  write to Trevor King, Drexel University, Physics Dept., 3141 Chestnut St.,
  Philadelphia PA 19104, USA.

 
  Plot simple (hopefully fast) strip charts.
  X bits from Xlib tutorial by Ch. Tronche (http://tronche.lri.fr:8000/)
  see http://tronche.com/gui/x/xlib/
  and http://tronche.com/gui/x/xlib-tutorial/2nd-program-anatomy.html
  
  Double buffering code following Mark Vojkovich
  http://www.xfree86.org/~mvojkovi/skull.tar.gz
 */

#include <pthread.h> /* to adapt to user input */
#include <stdlib.h>
#include <stdio.h>
#include <string.h> /* strlen */
//#include <plot.h> /* This is "plot.h", the public header file for GNU libplot,
//		     a shared library for 2-dimensional vector graphics */
#include <X11/Xthreads.h> /* since the callback thread and the main thread both call X */
#include <X11/Xlib.h>
#include <X11/Xatom.h> /* X datatypes, XA_STRING */
#include <X11/Xutil.h> /* XTextProperty, XParseGeometry */
#include "err_mac.h"
#include "stripchart.h"

//#define VERBOSE_LOCKING
//#define DEBUGGING_MODE /* Synchronize X */
#define USE_COLORS        /* Non-monochrome output */
#define USE_DOUBLE_BUFFER /* Write to a pixmap first to reduce screen redraw flicker */
#define ENABLE_AUTOSCALE  /* Calculate minX and minY from the data on mouse-clicks. */
#define WARN_WHEN_OOB     /* Generate some kind of warning output when out of bounds. */

#ifdef ENABLE_AUTOSCALE
#include <math.h> /* floor(), ceil() */
#endif

struct stripchart_struct {
  double minY;
  double maxY;
  double *data; /**< 1-D circular buffer of points */
  int points; /**< length of the buffer */
  int active_points; /**< while the buffer is still filling */
  int current_point; /**< our place in the buffer */
#ifdef ENABLE_AUTOSCALE
  double scale_res;
#endif

  /* display information for X */
  Display *display;
  int screen;
  int depth;
  Window rootwindow;
  Window window;
  XID drawable; /* either window or buffer, depending on use of double buffering */
  GC gc; /**< graphics context */
  char *title;
  int ulx;
  int uly;
  unsigned int width;
  unsigned int height;
  int point_radius; /* in pixels */
  int axes_on;
  XFontStruct *font;
  int blackpixel;
  int whitepixel;
#ifdef USE_COLORS
  XColor greenx, green; /* The X server and screen's respective takes on green */
#ifdef WARN_WHEN_OOB
  XColor redx, red;
#endif /* WARN_WHEN_OOB */
#endif /* USE_COLORS */
#ifdef USE_DOUBLE_BUFFER
  Pixmap buffer;
  GC copy_gc;
#endif
  /* pthread information so we can have a callback thread */
  pthread_mutex_t lock;
  int stop;
  /* done signal, so main thread can lock until callback wraps up? */

  /* derived quantites to avoid repead calculations */
  double xscale;    /* pxls between x values */
  double yscale;    /* pxls between per unit y */
};

/* Chart logic function */
static int scale(stripchart_t *chart);
static int autoscale(stripchart_t *chart);
static int draw(stripchart_t *chart);

/* X setup and drawing functions */
static int setup_X (stripchart_t *chart);
static int close_X (stripchart_t *chart);
static int set_title(stripchart_t *chart);
static int get_font(stripchart_t *chart);
#ifdef USE_COLORS
static int get_color(stripchart_t *chart, char *color_name, XColor *cx, XColor *c);
#endif
static int clear_window(stripchart_t *chart);
static int draw_point(stripchart_t *chart, int x, float y);
static int draw_axes(stripchart_t *chart);
static int flush_window(stripchart_t *chart);

/* Callback functions */
static int setup_callback_thread(stripchart_t *chart);
static void *callback_thread(void *arg /* chart */);
static int wakeup_callback_thread(stripchart_t *chart);

/* Synchronization functions */
static int lock_chart(stripchart_t *chart);
static int unlock_chart(stripchart_t *chart);

int stripchart_create(stripchart_t **pChart,
		      char *title,
		      char *geometry, /* eg "570x570+0+0" for upper left*/
		      double minY, double maxY,
		      double scale_res, /* only read if ENABLE_AUTOSCALE */
		      int points_shown_at_once,
		      int axes_on)
{
  stripchart_t *chart;
  int rc;

  AST(pChart != NULL);
  AST(*pChart == NULL);
  AST(geometry != NULL);
  AST(maxY > minY);
  AST(points_shown_at_once >= 1);
  
  chart = (stripchart_t *)malloc(sizeof(stripchart_t));
  *pChart = chart;
  AST(chart != NULL);
  
  chart->data = NULL;
  chart->title = NULL;
  chart->font = NULL;

  chart->minY = minY;
  chart->maxY = maxY;
#ifdef ENABLE_AUTOSCALE
  chart->scale_res = scale_res;
#endif
  chart->points = points_shown_at_once;
  chart->data = (double *)malloc(sizeof(double)*chart->points);
  AST(chart->data != NULL);
  chart->active_points = 0;
  chart->current_point = 0;

  chart->title = title;
  /* Set geometry defaults */
  chart->width=500;
  chart->height=200;
  chart->ulx=0;
  chart->uly=0;
  /* Parse geometry, rc is bitmask saying which were set */
  rc = XParseGeometry(geometry, &chart->ulx, &chart->uly,
		      &chart->width, &chart->height);
  chart->point_radius = 2;
  chart->axes_on = axes_on;
  CHK( scale(chart) );
  CHK( setup_X (chart) );
  CHK( setup_callback_thread (chart) );

  return SUCCESS;
}

/* needs a locked chart */
static int scale(stripchart_t *chart)
{
  /* xpxl = xscale * xindex , so (points-1) all the way on the right */
  chart->xscale = ((double)chart->width)/((double)chart->points-1.0);
  /* ypxl = yscale * (y-ymax), so ymax = 0, ymin = height */
  chart->yscale = ((double)chart->height)/(chart->minY - chart->maxY);
  return SUCCESS;
}

#ifdef ENABLE_AUTOSCALE
/* needs a locked chart */
static int autoscale(stripchart_t *chart)
{
  double dataMax, dataMin;
  int i;

  /* Ensure we have some data */
  if (chart->active_points == 0)
    return SUCCESS;
  /* Find the extreme data points */
  dataMax = dataMin = chart->data[0];
  for (i=1; i < chart->active_points; i++) {
    if (chart->data[i] > dataMax)
      dataMax = chart->data[i];
    else if (chart->data[i] < dataMin)
      dataMin = chart->data[i];
  }
  /* round out to units of scale_res */
  chart->maxY = ceil(dataMax/chart->scale_res)*chart->scale_res;
  chart->minY = floor(dataMin/chart->scale_res)*chart->scale_res;

  return SUCCESS;
}
#endif /* def ENABLE_AUTOSCALE */

/* needs a locked chart */
static int draw(stripchart_t *chart)
{
  int i, j;

  CHK( clear_window(chart) );
  /* dots */
  j = chart->current_point;
  for (i=0; i != chart->active_points; i++)
    CHK( draw_point(chart, i, chart->data[(j++)%chart->points]) );

  if (chart->axes_on)
    CHK( draw_axes(chart) );
  CHK( flush_window(chart) );

  return SUCCESS;
}

int stripchart_point(stripchart_t *chart, double value)
{
  int i, j;

  AST( chart != NULL );
  CHK( lock_chart(chart) );

  /* Add point */
  if (chart->active_points != chart->points) { /* no overwrite */
    chart->data[chart->active_points++] = value;
  } else { /* overwite an old point */
    chart->data[chart->current_point++] = value;
    chart->current_point %= chart->points;
  }
  CHK(draw(chart));
  CHK( unlock_chart(chart) );

  return SUCCESS;
}

int stripchart_destroy(stripchart_t **pChart)
{
  stripchart_t *chart;

  AST(pChart != NULL);
  chart = *pChart;
  CHK( lock_chart(chart) );
  chart = *pChart;
  *pChart = NULL;
  if (chart != NULL) {
    chart->stop = TRUE;
  }
  CHK( unlock_chart(chart) );
  CHK( wakeup_callback_thread(chart) );
  return SUCCESS;
}

/* called by the callback_thread
 * when it sees that it is stop time */
static int cleanup(stripchart_t *chart)
{
  AST(chart != NULL);
  if (chart->data != NULL)
    free(chart->data);
  close_X(chart);
  free(chart);
  return SUCCESS;
}

/* X setup and drawing functions */

static int setup_X (stripchart_t *chart)
{
  XEvent event;
  XSizeHints hints = {0};
#ifdef USE_DOUBLE_BUFFER
  XGCValues vals;
#endif

  /* Make X safe for threads */
  AST(XInitThreads() != 0);

  /* link to a display */
  chart->display = XOpenDisplay(NULL); /* argument? */
  if (chart->display == NULL) {
    fprintf(stderr, "Error connecting to %s\n", XDisplayName(NULL));
    AST(chart->display != NULL);
  }

#ifdef DEBUGGING_MODE
  XSynchronize(chart->display, 1);
#endif

  chart->screen = DefaultScreen(chart->display);
  chart->depth = DefaultDepth(chart->display, chart->screen);
  chart->rootwindow = DefaultRootWindow(chart->display);
  /* get some colors */
  chart->blackpixel = BlackPixel(chart->display, chart->screen);
  chart->whitepixel = WhitePixel(chart->display, chart->screen);

  /* create a window, window border != window manager's border */
  chart->window = XCreateSimpleWindow(chart->display, chart->rootwindow,
			  chart->ulx, chart->uly, chart->width, chart->height,
			  0 /* border width */,
			  chart->blackpixel /* border color */,
			  chart->blackpixel /* bg color */);
  AST( chart->window );
#ifdef USE_DOUBLE_BUFFER
  chart->buffer = XCreatePixmap(chart->display, chart->rootwindow,
				chart->width, chart->height, chart->depth);
  AST( chart->buffer );
  chart->drawable = chart->buffer;
#else
  chart->drawable = chart->window;
#endif

  CHK( set_title(chart) );
  //XMoveWindow(chart->display, chart->window, -1,-1);

  /* we want MapNotify events for
   * Structural events: window has appeared on the screen (been mapped),
   *  resized, moved, etc.
   * Exposure events: server lost information about the contents of a window
   * ButtonPress : mouse button clicked on window
   */
#ifdef ENABLE_AUTOSCALE
  XSelectInput(chart->display, chart->window,
	       StructureNotifyMask |
	       ExposureMask |
	       ButtonPressMask);
#else
  XSelectInput(chart->display, chart->window,
	       StructureNotifyMask |
	       ExposureMask);
#endif
  /* Map the window (make it appear on the screen) */
  XMapWindow(chart->display, chart->window);

  /* Create a "Graphics Context"
   * X is stateless, so the server doesn't remember attributes
   * like drawing color, the thickness of the lines and so on.
   * We have to pass these parameters to the server on each drawing request.
   * To avoid passing two dozens of parameters, many of them unchanged,
   * X uses an object called the Graphics Context, or GC for short.
   */
  chart->gc = XCreateGC(chart->display, chart->drawable, 0, NULL);
  
#ifdef USE_COLORS
  CHK( get_color(chart, "green", &chart->greenx, &chart->green) );
#ifdef WARN_WHEN_OOB
  CHK( get_color(chart, "red", &chart->redx, &chart->red) );
#endif /* WARN_WHEN_OOB */
#endif /* USE_COLORS */
#ifdef USE_DOUBLE_BUFFER
  /* turn off GraphicsExpose events for XCopyArea() requests */
  vals.graphics_exposures = False;
  chart->copy_gc = XCreateGC(chart->display, chart->window, GCGraphicsExposures, &vals);
#endif
  /* Prepare to draw with a white color */
  XSetForeground(chart->display, chart->gc, chart->whitepixel);
  /* leave other parameters at their default values */
  if (chart->axes_on)
    CHK( get_font(chart) );

  /* Wait for a MapNotify event */
  for (;;) {
    XNextEvent(chart->display, &event);
    if (event.type == MapNotify)
      break;
  }

  return SUCCESS;
}

static int close_X(stripchart_t *chart)
{
  XDestroyWindow(chart->display, chart->window);
  XFreePixmap(chart->display, chart->window);
  XFreeGC(chart->display, chart->gc);
#ifdef USE_DOUBLE_BUFFER
  XFreeGC(chart->display, chart->copy_gc);
#endif
  if (chart->font != NULL)
    XFreeFontInfo(NULL, chart->font, 1);
    /* free (char **names, XFontStruct *free_info, int actual_count) ? */
  XCloseDisplay(chart->display);
}

static int set_title(stripchart_t *chart)
{
  XTextProperty windowName;
  windowName.value    = (unsigned char *) chart->title;
  windowName.encoding = XA_STRING;
  windowName.format   = 8;
  windowName.nitems   = strlen((char *) windowName.value);

  /* request a title from the windows manager */
  XSetWMName(chart->display, chart->window, &windowName);

  return SUCCESS;
}

static int get_font(stripchart_t *chart)
{
  GContext gcont;

  gcont = XGContextFromGC(chart->gc);
  chart->font = XQueryFont(chart->display, gcont);

  return SUCCESS;
}

#ifdef USE_COLORS
static int get_color(stripchart_t *chart, char *color_name, XColor *cx, XColor *c)
{
  Colormap colormap;

  //AST( XLookupColor(chart->display, colormap, color_name,
  //		    NULL, &green_col) != 0 );
  
  if (DisplayPlanes(chart->display, chart->screen) != 1) {
    colormap = DefaultColormap(chart->display, chart->screen);
    if (XAllocNamedColor(chart->display, colormap, color_name, cx, c)) {
      /* use colors (how do I free the named color?) */
      /* cx is the idealized color, c is the best the current screen can do */
      //XSetForeground(chart->display, chart->gc2, color_screen.pixel);
      return SUCCESS;
    }
  }
  /* just use white */
  cx->pixel = chart->whitepixel;
  c->pixel = chart->whitepixel;

  return FAILURE;    
}
#endif /* def USE_COLORS */

static int clear_window(stripchart_t *chart)
{
  int x, y;
#ifdef USE_DOUBLE_BUFFER
  XSetForeground(chart->display, chart->gc, chart->blackpixel);
  XFillRectangle(chart->display, chart->buffer, chart->gc, 0, 0, chart->width, chart->height);
  XSetForeground(chart->display, chart->gc, chart->whitepixel);
  /* Send the requests to the server */
  //XFlush(chart->display);
#else
  XClearWindow(chart->display, chart->window);
#endif
  return SUCCESS;
}

static int flush_window(stripchart_t *chart)
{
#ifdef USE_DOUBLE_BUFFER
  /* XCopyArea(display, src, dest, gc, src_x, src_y, width, height,  dest_x, dest_y) */
  XCopyArea(chart->display, chart->buffer, chart->window, chart->copy_gc, 0, 0,
	    chart->width, chart->height, 0, 0);
  XFlush(chart->display);
#else
  XFlush(chart->display);
#endif
  return SUCCESS;
}

static int draw_point(stripchart_t *chart, int x, float y)
{
  int xpxl, ypxl;
  /* find the center of the circle in pixels */
  xpxl = (int) (chart->xscale * (float)x);
  ypxl = (int) (chart->yscale * (y - chart->maxY));

#ifdef WARN_WHEN_OOB
#ifdef USE_COLORS
  if (y > chart->maxY) {
    ypxl = 0; /* draw red point at the top of the screen */
    XSetForeground(chart->display, chart->gc, chart->red.pixel); 
  } else if (y < chart->minY) {
    ypxl = chart->height; /* draw red point at the bottom of the screen */
    XSetForeground(chart->display, chart->gc, chart->red.pixel); 
  }
#endif
#endif

  /* draw a circle, coords are (x,y) for upper left of bounding box
   * and the height and width of the bounding box.
   * Then start and stop angles in degrees * 64
   * XDrawArc and XFillArc take identical arguments, pick your favorite.
   */
  XDrawArc(chart->display, chart->drawable, chart->gc,
	   xpxl - chart->point_radius, ypxl - chart->point_radius,
	   2*chart->point_radius, 2*chart->point_radius,
	   0, 360*64);

  // XDrawLine(chart->display, chart->drawable, chart->gc, 10, 60, 180, 20);

#ifdef WARN_WHEN_OOB
#ifdef USE_COLORS
  XSetForeground(chart->display, chart->gc, chart->whitepixel); 
#endif
#endif

  return SUCCESS;
}

static int draw_axes(stripchart_t *chart)
{
  char buffer[80] = {0};
  
#ifdef USE_COLORS
  XSetForeground(chart->display, chart->gc, chart->green.pixel);
#endif
  /* lable y-max in upper left */
  sprintf(buffer, "%g", chart->maxY);
  XDrawString(chart->display, chart->drawable, chart->gc,
	      0,0+chart->font->ascent,buffer,strlen(buffer));

  /* label y-min in lower left */
  sprintf(buffer, "%g", chart->minY);
  XDrawString(chart->display, chart->drawable, chart->gc,
	      0,chart->height-chart->font->descent+1,
	      buffer,strlen(buffer));

  /* draw X axis */
  if (chart->maxY > 0 && chart->minY < 0) {
    int ypxl = (int) (-chart->yscale * chart->maxY);
    XDrawLine(chart->display, chart->drawable, chart->gc,
	      0, ypxl, chart->width, ypxl);
  }
#ifdef USE_COLORS
  XSetForeground(chart->display, chart->gc, chart->whitepixel);
#endif

  return SUCCESS;
}


static int handle_X_callbacks(stripchart_t *chart)
{
  XEvent event;

  /* Wait for a Expose event (which neccessitates redrawing. */
  for (;;) {
    XNextEvent(chart->display, &event);
    CHK( lock_chart(chart) );
    if (chart->stop == TRUE) {
      CHK( unlock_chart(chart) );
      goto done;
    }
    switch(event.type) {
    case Expose:
      CHK( draw(chart) );
      break;
    case ConfigureNotify:
      {
	XConfigureEvent *pe = (XConfigureEvent *)&event;
	chart->ulx = pe->x;
	chart->uly = pe->y;
	chart->width = pe->width;
	chart->height = pe->height;
      }
#ifdef USE_DOUBLE_BUFFER
      XFreePixmap(chart->display, chart->buffer);
      chart->buffer = XCreatePixmap(chart->display, chart->rootwindow,
				      chart->width, chart->height, chart->depth);
      AST( chart->buffer != 0 );
      chart->drawable = chart->buffer;
      //XClearWindow(chart->display, chart->window); /* ? */
#endif
      CHK( scale(chart) );     /* Map max and min to pixel values */
      CHK( draw(chart) );
      break;
#ifdef ENABLE_AUTOSCALE
    case ButtonPress:
      CHK( autoscale(chart) ); /* Calc new max and min values */
      CHK( scale(chart) );     /* Map max and min to pixel values */
      CHK( draw(chart) );
      break;
#endif
    default:
      break; /* ignore other events */
    } 
    CHK( unlock_chart(chart) );
  }
 done:
  return SUCCESS;
}

/* Callback functions */

static int setup_callback_thread(stripchart_t *chart)
{
  pthread_t tid;
  pthread_attr_t attr;

  CHK( pthread_mutex_init(&chart->lock, NULL) );
  CHK( pthread_attr_init(&attr) );
  CHK( pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED) );
  CHK( pthread_create(&tid, &attr, &callback_thread, (void *)chart) );
  CHK( pthread_attr_destroy(&attr) );

  return SUCCESS;
}

static void *callback_thread(void *arg /* chart */)
{
  stripchart_t *chart = (stripchart_t *)arg;
  handle_X_callbacks(chart);
  pthread_mutex_destroy(&chart->lock);
  cleanup(chart);
  return NULL;
}

static int wakeup_callback_thread(stripchart_t *chart)
{
  XEvent event;
  event.type = Expose;
  /* the callback thread is blocking while waiting for events,
     send it one to wake it, so it checks the stop value */
  XSendEvent (chart->display, chart->window,
	      FALSE, /* don't neet to propogate, we'll be listening */
	      ExposureMask,
	      &event);
  return SUCCESS;
}

/* Synchronization functions */

static int lock_chart(stripchart_t *chart)
{
#ifdef VERBOSE_LOCKING
  fprintf(stderr, "Thread %lu locking %p\n",
	  (unsigned long int)pthread_self(), &chart->lock);
#endif
  if (pthread_mutex_lock(&chart->lock) != 0) {
    perror("lock");
    AST(1==0);
  }
#ifdef VERBOSE_LOCKING
  fprintf(stderr, "Thread %lu locked %p\n",
	  (unsigned long int)pthread_self(), &chart->lock);
#endif
  return SUCCESS;
}

static int unlock_chart(stripchart_t *chart)
{
#ifdef VERBOSE_LOCKING
  fprintf(stderr, "Thread %lu unlocking %p\n",
	  (unsigned long int)pthread_self(), &chart->lock);
#endif
  if(pthread_mutex_unlock(&chart->lock) != 0) {
    perror("unlock");
    AST(1==0);
  }
#ifdef VERBOSE_LOCKING
  fprintf(stderr, "Thread %lu unlocked %p\n",
	  (unsigned long int)pthread_self(), &chart->lock);
#endif
  return SUCCESS;
}
